import numpy as np
import os
import os.path
import subprocess
import platform
import sys
import pkgutil
import re

# Adds a line to some *data file.
def add_keyword(keyword, filename:str, params=None):
    data_file = open(filename, "a")
    data_file.write(keyword)
    if params is not None:
        for p in params:
            data_file.write(" " + str(p))
    data_file.write("\n")

# Creates random coords file
def random_starting_coords(nweights):
    starting_coords = np.random.uniform(-5.0, 5.0,nweights)
    np.savetxt('coords',starting_coords[:,None], fmt='%.10f')
    
# change placeholders in some document
def fill_placeholders(fname:str, newfname:str, keyword:str, var):
    with open(fname, "r") as template:
        template_lines = template.readlines()
    with open(newfname, "w") as newtemp:
        for line in template_lines:
            edited_line = re.sub(keyword, str(var), line)
            newtemp.write(edited_line) 

# Run GMIN
def gmin():
    if os.path.isfile('data') and os.path.isfile('coords') and os.path.isfile('MLPdata'):
        # write output to file
        f = open("gmin_output.txt", "w")
        # execute GMIN binary as defined by setup.py for correct OS
        subprocess.call('GMIN', stdout=f)
    else:
        print('Data or oords file are missing')		

#run OPTIM
def optim(outfile='optim_output.txt'):
    if os.path.isfile('odata') and os.path.isfile('MLPdata'):
        # write output to file
        f = open(outfile, "w")
        # execute OPTIM binary as defined by setup.py for correct OS
        subprocess.call('OPTIM', stdout=f)
    else:
        print('odata file is missing')

#run PATHSAMPLE
def pathsample():
    if os.path.isfile('pathdata') and os.path.isfile('MLPdata'):
        # write output to file
        f = open("pathsample_output.txt", "w")
        # execute PS binary as defined by setup.py for correct OS
        subprocess.call('PATHSAMPLE', stdout=f)
    else:
        print('pathdata file is missing')


#initiate PATHSAMPLE database from lowest
def initpdb(dof): 
    lowest_lines=dof+2
    nlines=sum(1 for line in open('lowest'))
    nmin=int(nlines/lowest_lines)
    mlp_lines=sum(1 for line in open('MLPdata'))
    # edit dumpdata file
    fill_placeholders("odata.dumpdata.template","odata.dumpdata","MLP_LINES",mlp_lines)
    # extract minima one by one and write to pathsample db afterwards
    cdir = os.getcwd()
    for lmin in range(nmin):
        shutil.copyfile(cdir+'/odata.dumpdata',(cdir+'/odata'))
        head_lines=(lmin+1)*lowest_lines
        open("odata", "a").writelines([l for n,l in enumerate(open("lowest").readlines()) if n < head_lines if n >= (head_lines-dof)])
        # run OPTIM now
        optim()
        # reformat OPTIM output
        open("min.data.info.all","a").writelines([l for l in open("min.data.info").readlines()])
        lmin+=1
    # edit pathdata file
    fill_placeholders("pathdata.template.ADDMIN","pathdata","MLP_LINES",mlp_lines)
    # run PATHSAMPLE
    pathsample()
    open("min.A",'w').writelines("1\n1")
    open("min.B",'w').writelines("1\n2")
    os.remove('min.data.info')
    os.remove('min.data.info.all')


def extendpdb(curr_min=1):
    mlp_lines=sum(1 for line in open('MLPdata'))
    # first write pathdata file
    fill_placeholders("pathdata.template.NEWCONNECTIONS","pathdata","MLP_LINES",mlp_lines)
    fill_placeholders("pathdata","pathdata","CURR_MIN",curr_min)
    # next do both odata files
    fill_placeholders("odata.tspath.template","odata.tspath","MLP_LINES",mlp_lines)
    fill_placeholders("odata.connect.template","odata.connect","MLP_LINES",mlp_lines)
    # finally just run PATHSAMPLE
    pathsample()
    os.remove('path.info.*')
    os.remove('submit_*')
    

def auc():
    mlp_lines=sum(1 for line in open('MLPdata'))
    fill_placeholders("pathdata.template.EXTRACTMIN","pathdata","MLP_LINES",mlp_lines)
    nmin=sum(1 for line in open('min.data'))
    fill_placeholders("odata.AUC.template","odata.AUC","MLP_LINES",mlp_lines)
    fill_placeholders("odata.AUC","odata","START_NO",str(1))
    if os.path.isfile('extractedmin.all'): os.remove('extractedmin.all')
    if os.path.isfile('extractedmin'): os.remove('extractedmin')
    for ind in range(nmin):
        minind=ind+1
        fill_placeholders("pathdata","pathdata","EXTRACT_NO",minind)
        pathsample()
        if minind == 1:
            open("odata", "a").writelines([l for l in open("extractedmin").readlines()])
        else:
            open("extractedmin.all","a").writelines([l for l in open("extractedmin").readlines()])
    open("extractedmin", "w").writelines([l for l in open("extractedmin.all").readlines()])   
    os.remove('extractedmin.all')
    optim(outfile='auc_logfile')
    # extract AUCs from optim logfile
    with open("auc_logfile","r") as f:
        auclist=[]
        for line in f:
            if line.startswith("energy, AUC="):
                auclist.append(line[8:])
    aucs = [elem.split()[2] for elem in auclist]
    aucs = list(map(float,aucs))
    np.savetxt('AUCs_unsorted',aucs,fmt='%.10f')
    sorted_aucs = np.sort(aucs)[::-1]
    np.savetxt('AUCs_sorted',sorted_aucs,fmt='%.10f')
    best_auc = np.max(aucs)
    return(best_auc) 















from .functions import add_keyword
import numpy as np
import os

# data file for GMIN
def make_gmin_data(nin,nhidden,nout,norm=True,ndata="MLP_LINES",l2lambda="0.00001",nsteps=100000,f_outname='data.template'):
    fname = f_outname
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    add_keyword("SLOPPYCONV", fname, ["1.0D-6"])
    add_keyword("TIGHTCONV", fname, ["1.0D-9"])
    if norm==True:
        add_keyword("MLPNORM",fname)
    add_keyword("SAVE", fname,[1000])
    add_keyword("UPDATES", fname,[500])
    add_keyword("MAXIT", fname,[5000,5000])
    add_keyword("EDIFF", fname, ["1.0D-04"])
    add_keyword("STEPS", fname, [nsteps, 1.0])
    add_keyword("STEP", fname, ["1.0"])
    add_keyword("TEMPERATURE", fname, ["2.0"])
    add_keyword("DUMPINT", fname,[100])
    add_keyword("FIXBOTH",fname)
    add_keyword("MLPVB3", fname,[nin,1,nhidden,nout,ndata,l2lambda])

# odata dumpdata file to generate pathsample database
def make_odata_dumpdata(nin,nhidden,nout,norm=True,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'odata.dumpdata.template'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    if norm==True:
        add_keyword("MLPNORM",fname)
    add_keyword("SEARCH",fname,[0])
    add_keyword("DUMPDATA",fname)
    add_keyword("ENDHESS",fname)
    add_keyword("MAXBFGS",fname,["0.1","0.1"])
    add_keyword("UPDATES",fname,[1000,1000])
    add_keyword("BFGSMIN",fname,["1.0D-10"])
    add_keyword("STEPS",fname,[10000])
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword("VARIABLES",fname)

# write dumped minima from lowest to pathsample db
def make_pathdata_addmin(nin,nhidden,nout,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'pathdata.template.ADDMIN'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword("ALLTS",fname)    
    add_keyword('EVCUT',fname,['1.0D-7'])
    add_keyword('COPYFILES',fname,['MLPdata'])
    add_keyword('EDIFFTOL',fname,['1.0D-4'])
    add_keyword('GEOMDIFFTOL',fname,['1.0D2'])
    add_keyword('EXEC',fname,['OPTIM'])
    add_keyword("TEMPERATURE", fname, ["0.1"])
    add_keyword('PAIRLIST',fname,[1])
    add_keyword('ADDMIN',fname,['min.data.info.all'])

# extend pathsample db with one of the 4 keywords (3 commented out)
def make_pathdata_newconnections(nin,nhidden,nout,ndata="MLP_LINES",l2lambda="0.00001",curr_min=1):
    fname = 'pathdata.template.NEWCONNECTIONS'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword("ALLTS",fname)
    add_keyword('EVCUT',fname,['1.0D-7'])
    add_keyword('COPYFILES',fname,['MLPdata'])
    add_keyword('EDIFFTOL',fname,['1.0D-4'])
    add_keyword('GEOMDIFFTOL',fname,['1.0D2'])
    add_keyword('EXEC',fname,['OPTIM'])
    add_keyword("TEMPERATURE", fname, ["0.1"])
    add_keyword('CYCLES',fname,[5000])
    add_keyword('PAIRLIST',fname,[1])
    add_keyword('NEWCONNECTIONS',fname,[10000000,1000,curr_min])    
    add_keyword('comment UNTRAP',fname,['0.001','0.001'])
    add_keyword('comment CONNECTUNC',fname,['EREF','2'])
    add_keyword('comment CONNECTREGION',fname,['1','58','1.0D100'])
    add_keyword('SEED',fname,[str(np.random.randint(100000))])

# for pathsample db extension
def make_odata_connect(nin,nhidden,nout,norm=True,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'odata.connect.template'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    if norm==True:
        add_keyword("MLPNORM",fname)
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword('STEPMIN',fname,[100])
    add_keyword('NEWCONNECT',fname,[100,1,10.0,20.0,50,5.0,0.025])
    add_keyword('MAXERISE',fname,['1.0D-6','1.0D-3'])
    add_keyword('REOPTIMISEENDPOINTS',fname)
    add_keyword('NEWNEB',fname,[100,350,0.0001])
    add_keyword('DIJKSTRA',fname,['EXP'])
    add_keyword('MAXBFGS',fname,[0.1,0.1])
    add_keyword('EDIFFTOL',fname,['1.0D-4'])
    add_keyword('GEOMDIFFTOL',fname,['1.0D2'])
    add_keyword('DUMPALLPATHS',fname)
    add_keyword('MAXSTEP',fname,[0.01])
    add_keyword('MAXMAX',fname,[0.2])
    add_keyword('TRAD',fname,[1.0])
    add_keyword('NEBK',fname,[50.0])
    add_keyword('BFGSTS',fname,[5000,1,10,0.002])
    add_keyword('NOIT',fname)
    add_keyword('UPDATES',fname,[10,10])
    add_keyword('BFGSMIN',fname,['1.0D-9'])
    add_keyword('CONVERGE',fname,['1.0','1.0D-9'])
    add_keyword('PUSHOFF',fname,[0.2])
    add_keyword('STEPS',fname,[1000])
    add_keyword('BFGSSTEPS',fname,[50000])
    add_keyword('VARIABLES',fname)

# for pathsample db extension
def make_odata_tspath(nin,nhidden,nout,norm=True,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'odata.tspath.template'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    if norm==True:
        add_keyword("MLPNORM",fname)
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword('DUMPALLPATHS',fname)
    add_keyword('BFGSMIN',fname,['1.0D-9'])
    add_keyword('CONVERGE',fname,['1.0','1.0D-9'])
    add_keyword('EDIFFTOL',fname,['1.0D-4'])
    add_keyword('GEOMDIFFTOL',fname,['1.0D2'])
    add_keyword('BFGSTS',fname,[5000,1,10,0.002])
    add_keyword('NOIT',fname)
    add_keyword('PATH',fname,[3,0.0])
    add_keyword('SHIFT',fname,['1.0D10'])
    add_keyword('MAXSTEP',fname,[0.02])
    add_keyword('MAXMAX',fname,[0.04])
    add_keyword('TRAD',fname,['1.0D0'])
    add_keyword('STEPS',fname,[1000])
    add_keyword('UPDATES',fname,[10,10])
    add_keyword('BFGSSTEPS',fname,[50000])
    add_keyword('VARIABLES',fname)


def make_odata_auc(nin,nhidden,nout,norm=True,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'odata.AUC.template'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    if norm==True:
        add_keyword("MLPNORM",fname)
    add_keyword('MLPPROB',fname)
    add_keyword('BFGSMIN',fname,['1.0D100'])
    add_keyword('MULTIJOB',fname,['extractedmin'])
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword('VARIABLES',fname)


def make_pathdata_extractmin(nin,nhidden,nout,ndata="MLP_LINES",l2lambda="0.00001"):
    fname = 'pathdata.template.EXTRACTMIN'
    # remove old file
    if os.path.isfile(fname): os.remove(fname)
    # write new file
    add_keyword("MLPVB3",fname,[nin,1,nhidden,nout,ndata,l2lambda])
    add_keyword("ALLTS",fname)
    add_keyword('EVCUT',fname,['1.0D-7'])
    add_keyword('COPYFILES',fname,['MLPdata'])
    add_keyword('EDIFFTOL',fname,['1.0D-4'])
    add_keyword('GEOMDIFFTOL',fname,['1.0D2'])
    add_keyword('EXEC',fname,['OPTIM'])
    add_keyword("TEMPERATURE", fname, ["0.1"])
    add_keyword('PAIRLIST',fname,[1])
    add_keyword('EXTRACTMIN',fname,['EXTRACT_NO'])

def make_dinfo():
    fname='dinfo.template'
    add_keyword("DELTA",fname,['ENERGY_DIFFERENCE'])
    add_keyword("FIRST",fname,['HIGHEST_ENERGY'])
    add_keyword("comment MAXTSENERGY",fname,[0.395])
    add_keyword("CENTREGMIN",fname)
    add_keyword("comment CONNECTMIN",fname,[1])
    add_keyword("LEVELS",fname,['NO_OF_LEVELS'])
    add_keyword("MINIMA",fname,['min.data'])
    add_keyword("TS",fname,['ts.data'])
    add_keyword("comment IDENTIFY",fname)
    add_keyword("comment LOWEST",fname,[5000])
    add_keyword("comment MONOTONIC",fname)
    add_keyword("comment NCONNMIN",fname,[0])
    add_keyword("comment SCALEBAR",fname,[1])
    add_keyword("comment LABELFORMAT",fname,['E12.4'])
    add_keyword("comment TRMIN",fname,['2','NO_OF_MIN','AUCs2','AUCs1']) 
    add_keyword("comment TRMIN",fname,['1','NO_OF_MIN','AUCs1'])
    add_keyword("comment CHOOSECOLOURS",fname)










import math
import os
import os.path
import re
import subprocess

def maketree():
    with open("dinfo.template", "r") as template:
        template_lines = template.readlines()

    min_energies = []
    try:
        with open("min.data", "r") as f:
            for line in f:
                min_energies.append(float(line.split()[0]))
    except IOError:
        print(working_dir + ": min.data does not exist")
        os.chdir("..")
    min_count = len(min_energies)

    ts_energies = []
    try:
        with open("ts.data", "r") as f:
            for line in f:
                ts_energies.append(float(line.split()[0]))
    except IOError:
        print(working_dir + ": ts.data does not exist")
        os.chdir("..")

    if len(min_energies) == 1 and len(ts_energies) == 0:
        print(working_dir + ": database has 1 minimum and 0 ts")
        os.chdir("..")

    highest_energy = max(ts_energies)
    levels = 30
    delta = (highest_energy - min(min_energies)) / (levels - 1)
    # Round up to one significant figure (assumes 0 < delta < 1)
    mag = 0
    while (delta < 1):
        delta *= 10
        mag += 1
    delta = math.ceil(delta)
    while (mag > 0):
        delta /= 10
        mag -= 1

    with open("dinfo", "w") as dinfo:
        for line in template_lines:
            edited_line = re.sub("HIGHEST_ENERGY", str(highest_energy), line)
            edited_line = re.sub("ENERGY_DIFFERENCE", str(delta), edited_line)
            edited_line = re.sub("NO_OF_LEVELS", str(levels), edited_line)
            edited_line = re.sub("NO_OF_MIN", str(min_count), edited_line)
            dinfo.write(edited_line)

    subprocess.call(["disconnectionDPS"])

    # Substitute the scale epsilon
    try:
        with open("tree.ps", "r") as tree:
            tree_lines = tree.readlines()
        with open("tree.ps", "w") as tree:
            for line in tree_lines:
                edited_line = re.sub("epsilon", str(delta), line)
                tree.write(edited_line)
    except (IOError):
      print("tree.ps does not exist")




import setuptools
import platform


# Determine the correct platform for binaries
system = platform.system()
#arch, _ = platform.architecture()
if system == 'Linux':
    gmin = 'bin/linux/GMIN'
    optim = 'bin/linux/OPTIM'
    pathsample = 'bin/linux/PATHSAMPLE'
    disconnectionDPS = 'bin/linux/disconnectionDPS'
if system == 'Windows':
    gmin = 'bin/win/GMIN.exe'
    optim = 'bin/win/OPTIM.exe'
    pathsample = 'bin/win/PATHSAMPLE.exe'
    disconnectionDPS = 'bin/win/disconnectionDPS.exe'
if system == 'Darwin':
    gmin = 'bin/mac/GMIN'
    optim = 'bin/mac/OPTIM'
    pathsample = 'bin/mac/PATHSAMPLE'
    disconnectionDPS = 'bin/mac/disconnectionDPS'

print(gmin)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pylfl",
    version="0.1.3",
    author="Max Niroomand",
    author_email="mpn26@cam.ac.uk",
    description="A package to survey LFLs in ML models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/orinxam/pylfl",
    project_urls={
        "Bug Tracker": "https://github.com/orinxam/pylfl/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    data_files = [('bin', [gmin,optim,pathsample,disconnectionDPS])],
)



